/*
** Copyright (C) 2008-2009 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

#include "silk.h"

RCSIDENT("$SiLK: flowrate.c 13660 2009-02-27 16:33:53Z mthomas $");

#include "rwpack.h"
#include "dynlib.h"
#include <float.h>

/*
**  flowrate.c
**
**    Plug-in to allow filtering, sorting, counting, and printing of
**    the following values:
**
**    -- packets-per-second
**    -- bytes-per-second
**    -- bytes-per-packet (not for rwfilter; it already exists)
**    -- payload-bytes
**    -- payload-bytes-per-second
**
**    Mark Thomas
**    July 2008
*/


/* EXPORTED FUNCTIONS */

/* The dynamic library glue code calls these */
int dynlib_api_version(void);
int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType);
int configure(dynlibInfoStruct_t *dlISP, void *data);
void teardown(dynlibSymbolId UNUSED(appType));
void optionsUsage(dynlibSymbolId appType, FILE *fh);
int filter(rwRec *rwrec);
int cut(unsigned int field, char *out, size_t len_out, rwRec *rwrec);
int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec);
int uniq(
    unsigned int    field,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);


/* DEFINES AND TYPEDEFS */

/* identifiers for the fields */
#define PCKTS_PER_SEC       1
#define BYTES_PER_SEC       2
#define BYTES_PER_PACKET    3
#define PAYLOAD_BYTES       4
#define PAYLOAD_RATE        5

/* the size of the binary value used in rwsort and rwuniq */
#define RATE_BINARY_SIZE  sizeof(uint64_t)

/* preferred width of textual columns */
#define RATE_TEXT_WIDTH   16

/* in rwsort and rwuniq, we encode the value as a uint64_t in
 * big-endian format.  To get the uint64_t, we multiply the double
 * value by this precision. */
#define PRECISION 1e4

#define DOUBLE_TO_UINT64(v) ((uint64_t)((v) * PRECISION))
#define UINT64_TO_DOUBLE(v) ((double)(v) / PRECISION)

/* how to calculate the packets per second.  rwRecGetElapsed() returns
 * the value in milliseconds.  If the elapsed time is 0, just return
 * the packet count. */
#define PCKT_RATE_DOUBLE(r)                                             \
    ((rwRecGetElapsed(r) == 0)                                          \
     ? rwRecGetPkts(r)                                                  \
     : ((double)(rwRecGetPkts(r)) * 1000.0 / (double)(rwRecGetElapsed(r))))

/* as above, but for bytes-per-second */
#define BYTE_RATE_DOUBLE(r)                                             \
    ((rwRecGetElapsed(r) == 0)                                          \
     ? rwRecGetBytes(r)                                                 \
     : ((double)(rwRecGetBytes(r)) * 1000.0 / (double)(rwRecGetElapsed(r))))

/* as above, for the payload-rate */
#define PAYLOAD_RATE_DOUBLE(r)                                          \
    ((rwRecGetElapsed(r) == 0)                                          \
     ? getPayload(r)                                                    \
     : ((double)getPayload(r) * 1000.0 / (double)(rwRecGetElapsed(r))))

/* how to calculate bytes-per-packet */
#define BYTES_PER_PACKET_DOUBLE(r)                              \
    ((double)rwRecGetBytes(r) / (double)rwRecGetPkts(r))


/* structures to hold min and max values */
typedef struct double_range_st {
    double      min;
    double      max;
    unsigned    is_active :1;
} double_range_t;

typedef struct u64_range_st {
    uint64_t    min;
    uint64_t    max;
    unsigned    is_active :1;
} u64_range_t;


/* LOCAL VARIABLES */

static const char *plugin_name = "flowrate";

/* for filtering, pass records whose packets-per-second,
 * bytes-per-second, payload-bytes, and payload-bytes-per-second
 * values fall within these ranges. */
static double_range_t pckt_rate = {0, DBL_MAX, 0};
static double_range_t byte_rate = {0, DBL_MAX, 0};
static double_range_t payload_rate = {0, DBL_MAX, 0};
static u64_range_t payload_bytes = {0, UINT64_MAX, 0};

typedef enum plugin_options_en {
    OPT_PACKETS_PER_SECOND,
    OPT_BYTES_PER_SECOND,
    OPT_PAYLOAD_BYTES,
    OPT_PAYLOAD_RATE
} plugin_options_enum;

static struct option plugin_options[] = {
    {"packets-per-second",  REQUIRED_ARG, 0, OPT_PACKETS_PER_SECOND},
    {"bytes-per-second",    REQUIRED_ARG, 0, OPT_BYTES_PER_SECOND},
    {"payload-bytes",       REQUIRED_ARG, 0, OPT_PAYLOAD_BYTES},
    {"payload-rate",        REQUIRED_ARG, 0, OPT_PAYLOAD_RATE},
    {0, 0, 0, 0}            /* sentinel */
};

static const char *plugin_help[] = {
    "Packets-per-second is within decimal range X-Y.",
    "Bytes-per-second is within decimal range X-Y.",
    "Payload-byte count is within integer range X-Y.",
    "Payload-bytes-per-second is within decimal range X-Y.",
    NULL
};

/* fields for rwcut, rwuniq, etc */
static sk_stringmap_entry_t plugin_fields[] = {
    {"pckts/sec",       PCKTS_PER_SEC},
    {"bytes/sec",       BYTES_PER_SEC},
    {"bytes/packet",    BYTES_PER_PACKET},
    {"payload-bytes",   PAYLOAD_BYTES},
    {"payload-rate",    PAYLOAD_RATE},
    {NULL,              UINT32_MAX}     /* sentinel */
};


/* LOCAL FUNCTION PROTOTYPES */

static int optionsHandler(clientData cData, int opt_index, char *opt_arg);
static int parseDecimalRange(
    double_range_t *range,
    const char     *range_string,
    int             opt_index);


/* FUNCTION DEFINITIONS */

/*
 *  int dynlib_api_version(void);
 *
 *    Return the dynlib API version that this plugin was compiled
 *    against.
 */
int dynlib_api_version(void)
{
    return DYNLIB_API_VERSION;
}


/*
 * int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType)
 *
 * Called by dynlib interface code to set up the plugin.  This routine
 * should set up options handling if required.  Arguments: pointer to
 * the dynamic library interface structure, application type that is
 * using this plugin.  Returns DYNLIB_FAILED if processing fails,
 * DYNLIB_WONTPROCESS if the application should do normal output, and
 * DYNLIB_WILLPROCESS if this plugin takes over output.
 */
int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType)
{
    /* Make the plugin use the application's context, so we can add
     * options to main app. */
    skAppContextSet(dynlibGetAppContext(dlISP));

    switch (appType) {
      case DYNLIB_SHAR_FILTER:
        if (skOptionsRegister(plugin_options, &optionsHandler,
                              (clientData)dlISP))
        {
            skAppPrintErr("Unable to register plugin options.");
            return DYNLIB_FAILED;
        }
        break;

      case DYNLIB_CUT:
      case DYNLIB_SORT:
      case DYNLIB_UNIQ:
        /* no options.  the fields will get added in the 'configure'
         * function. */
        break;

      default:
        skAppPrintErr("Cannot use %s plug-in with %s application",
                      plugin_name, skAppName());
        return DYNLIB_FAILED;
    }

    return DYNLIB_WONTPROCESS;
}


/*
 *  int configure(dynlibInfoStruct_t *dlISP, void *data)
 *
 *    Perform additional configuration of the plug-in prior to
 *    processing the options.
 */
int configure(dynlibInfoStruct_t *dlISP, void *data)
{
    const dynlibSymbolId appType = dynlibGetAppType(dlISP);

    switch(appType) {
      case DYNLIB_CUT:
      case DYNLIB_SORT:
      case DYNLIB_UNIQ:
        /* for these apps, data is a pointer into which we should copy
         * the address of the fields this plug-in provides to the
         * application */
        if (data == NULL) {
            return -1;
        } else {
            sk_stringmap_entry_t **fields_ptr = (sk_stringmap_entry_t **)data;
            *fields_ptr = plugin_fields;
        }
        break;

      default:
        /* nothing to do */
        break;
    }

    return 0;
}


/*
 *  void teardown(dynlibSymbolId appType);
 *
 *    Called by dynlib interface code to tear down this plugin.
 */
void teardown(dynlibSymbolId UNUSED(appType))
{
    return;
}


/*
 * void optionsUsage(dynlibSymbolId UNUSED(appType), FILE *fh);
 *
 * Called by the dynlib interface to allow this plugin to print the
 * options it accepts.
 */
void optionsUsage(dynlibSymbolId appType, FILE *fh)
{
    int i;

    if (appType != DYNLIB_SHAR_FILTER) {
        return;
    }
    for (i = 0; plugin_options[i].name; i++) {
        fprintf(fh, "--%s %s. %s\n", plugin_options[i].name,
                SK_OPTION_HAS_ARG(plugin_options[i]), plugin_help[i]);
    }
}


/*
 * int optionsHandler(clientData cData, int opt_index, char *opt_arg);
 *
 * Handles options for the plugin.  cData is the dynlib ISP (which can
 * be used to find which application we are, among other things.)
 * opt_index is the enum passed in with the options descriptions.  opt_arg
 * is the argument, or NULL if no argument was given.  Returns 0 on
 * success, or 1 if there was a problem.
 */
static int optionsHandler(clientData cData, int opt_index, char *opt_arg)
{
    const dynlibSymbolId appType =dynlibGetAppType((dynlibInfoStruct_t*)cData);
    int rv;

    if (DYNLIB_SHAR_FILTER != appType) {
        skAppPrintErr("The %s plug-in does not support this application",
                      plugin_name);
        return 1;
    }

    switch ((plugin_options_enum)opt_index) {
      case OPT_PACKETS_PER_SECOND:
        if (parseDecimalRange(&pckt_rate, opt_arg, opt_index)) {
            return 1;
        }
        break;

      case OPT_BYTES_PER_SECOND:
        if (parseDecimalRange(&byte_rate, opt_arg, opt_index)) {
            return 1;
        }
        break;

      case OPT_PAYLOAD_RATE:
        if (parseDecimalRange(&payload_rate, opt_arg, opt_index)) {
            return 1;
        }
        break;

      case OPT_PAYLOAD_BYTES:
        rv = skStringParseRange64(&payload_bytes.min, &payload_bytes.max,
                                  opt_arg, 0, 0, SKUTILS_RANGE_SINGLE_OPEN);
        if (rv) {
            skAppPrintErr("Invalid %s '%s': %s",
                          plugin_options[opt_index].name, opt_arg,
                          skStringParseStrerror(rv));
            return 1;
        }
        payload_bytes.is_active = 1;
        break;
    }

    return 0;
}


/*
 *  status = parseDecimalRange(&range, opt_arg, opt_index);
 *
 *    Parse the string in 'opt_arg' as the range of two decimal
 *    values, and put the bounds into 'range'.  Return 0 on success.
 *    If an error occurs, use 'opt_index' to report the name of the
 *    option that caused the problem and return non-zero.
 */
static int parseDecimalRange(
    double_range_t *range,
    const char     *range_string,
    int             opt_index)
{
    int rv;

    range->is_active = 1;

    rv = skStringParseDoubleRange(&range->min, &range->max,
                                  range_string, 0.0, 0.0,
                                  SKUTILS_RANGE_SINGLE_OPEN);
    if (rv) {
        /* error */
        skAppPrintErr("Invalid %s '%s': %s",
                      plugin_options[opt_index].name, range_string,
                      skStringParseStrerror(rv));
        return 1;
    }

    return 0;
}


/*
 *  payload = getPayload(rwrec);
 *
 *    Compute the bytes of payload contained in 'rwrec' by multiplying
 *    the number of packets by the packet overhead and subtracting
 *    that from the byte count.  Return 0 if that value would be
 *    negative.
 *
 *    This function assumes minimal packet headers---that is, there
 *    are no options in the packets.  For TCP, assumes there are no
 *    TCP timestamps in the packet.  The returned value will be the
 *    maximum possible bytes of payload.
 */
static uint64_t getPayload(const rwRec *rwrec)
{
    uint64_t overhead;

#if SK_ENABLE_IPV6
    if (rwRecIsIPv6(rwrec)) {
        /* IPv6 IP-header header with no options is 40 bytes */
        switch (rwRecGetProto(rwrec)) {
          case IPPROTO_TCP:
            /* TCP header is 20 (no TCP timestamps) */
            overhead = rwRecGetPkts(rwrec) * 60;
            break;
          case IPPROTO_UDP:
            /* UDP header is 8 bytes */
            overhead = rwRecGetPkts(rwrec) * 48;
            break;
          default:
            overhead = rwRecGetPkts(rwrec) * 40;
            break;
        }
    } else
#endif  /* SK_ENABLE_IPV6 */
    {
        /* IPv4 IP-header header with no options is 20 bytes */
        switch (rwRecGetProto(rwrec)) {
          case IPPROTO_TCP:
            overhead = rwRecGetPkts(rwrec) * 40;
            break;
          case IPPROTO_UDP:
            overhead = rwRecGetPkts(rwrec) * 28;
            break;
          default:
            overhead = rwRecGetPkts(rwrec) * 20;
            break;
        }
    }

    if (overhead > rwRecGetBytes(rwrec)) {
        return 0;
    }

    return (rwRecGetBytes(rwrec) - overhead);
}


/*
 *  int filter(rwRec *rwrec);
 *
 *    The function actually used to implement filtering for the
 *    plugin.  Returns 0 if the record passes the filter, 1 if it
 *    fails the filter.
 */
int filter(rwRec *rwrec)
{
    uint64_t payload;
    double rate;

    /* filter by packets-per-second */
    if (pckt_rate.is_active) {
        rate = PCKT_RATE_DOUBLE(rwrec);
        if (rate < pckt_rate.min || rate > pckt_rate.max) {
            /* failed */
            return 1;
        }
    }

    /* filter by bytes-per-second */
    if (byte_rate.is_active) {
        rate = BYTE_RATE_DOUBLE(rwrec);
        if (rate < byte_rate.min || rate > byte_rate.max) {
            /* failed */
            return 1;
        }
    }

    /* filter by payload-bytes */
    if (payload_bytes.is_active) {
        payload = getPayload(rwrec);
        if (payload < payload_bytes.min || payload > payload_bytes.max) {
            /* failed */
            return 1;
        }
    }

    /* filter by payload-rate */
    if (payload_rate.is_active) {
        rate = PAYLOAD_RATE_DOUBLE(rwrec);
        if (rate < payload_rate.min || rate > payload_rate.max) {
            /* failed */
            return 1;
        }
    }

    return 0;
}


/*
 *  status = getTitle(text_val, text_len, &index);
 *
 *    For rwcut, rwsort, rwuniq, write the title of the column
 *    referenced by '*index' into the buffer 'text_val', a buffer of
 *    'text_len' characters.
 */
static int getTitle(
    char       *text_value,
    size_t      text_size,
    void       *idx)
{
    assert(*((unsigned int*)(idx)) >= 1);
    assert(*((unsigned int*)(idx)) <= 5);

    return snprintf(text_value, text_size, "%s",
                    plugin_fields[(*((unsigned int*)(idx)) - 1)].name);
}


/*
 *  status = recToText(rwrec, text_val, text_len, &index);
 *
 *    Given the SiLK Flow record 'rwrec', compute the flow-rate ratio
 *    specified by '*index', and write a textual representation of
 *    that value into 'text_val', a buffer of 'text_len' characters.
 */
static int recToText(
    const rwRec    *rwrec,
    char           *text_value,
    size_t          text_size,
    void           *idx)
{
    switch (*((unsigned int*)(idx))) {
      case PCKTS_PER_SEC:
        return snprintf(text_value, text_size, "%.3f",
                        PCKT_RATE_DOUBLE(rwrec));
      case BYTES_PER_SEC:
        return snprintf(text_value, text_size, "%.3f",
                        BYTE_RATE_DOUBLE(rwrec));
      case BYTES_PER_PACKET:
        return snprintf(text_value, text_size, "%.3f",
                        BYTES_PER_PACKET_DOUBLE(rwrec));
      case PAYLOAD_BYTES:
        return snprintf(text_value, text_size, ("%" PRIu64),
                        getPayload(rwrec));
      case PAYLOAD_RATE:
        return snprintf(text_value, text_size, "%.3f",
                        PAYLOAD_RATE_DOUBLE(rwrec));
    }
    return -1;
}


/*
 *  status = recToBin(rwrec, bin_val, &index);
 *
 *    Given the SiLK Flow record 'rwrec', compute the flow-rate ratio
 *    specified by '*index', and write a binary representation of
 *    that value into 'bin_val'.
 */
static int recToBin(const rwRec *rwrec, uint8_t *bin_value, void *idx)
{
    uint64_t val_u64;

    switch (*((unsigned int*)(idx))) {
      case PCKTS_PER_SEC:
        val_u64 = DOUBLE_TO_UINT64(PCKT_RATE_DOUBLE(rwrec));
        break;
      case BYTES_PER_SEC:
        val_u64 = DOUBLE_TO_UINT64(BYTE_RATE_DOUBLE(rwrec));
        break;
      case BYTES_PER_PACKET:
        val_u64 = DOUBLE_TO_UINT64(BYTES_PER_PACKET_DOUBLE(rwrec));
        break;
      case PAYLOAD_BYTES:
        val_u64 = getPayload(rwrec);
        break;
      case PAYLOAD_RATE:
        val_u64 = DOUBLE_TO_UINT64(PAYLOAD_RATE_DOUBLE(rwrec));
        break;
      default:
        return -1;
    }

    val_u64 = hton64(val_u64);
    memcpy(bin_value, &val_u64, RATE_BINARY_SIZE);

    /* return width of binary value */
    return (int)RATE_BINARY_SIZE;
}


/*
 *  status = binToText(bin_val, text_val, text_len, &index);
 *
 *    Given the buffer 'bin_val' which was filled by calling
 *    recToBin(), write a textual representation of that value into
 *    'text_val', a buffer of 'text_len' characters.
 */
static int binToText(
    const uint8_t  *bin_value,
    char           *text_value,
    size_t          text_size,
    void    UNUSED(*idx))
{
    uint64_t val_u64;

    switch (*((unsigned int*)(idx))) {
      case PCKTS_PER_SEC:
      case BYTES_PER_SEC:
      case BYTES_PER_PACKET:
      case PAYLOAD_RATE:
        memcpy(&val_u64, bin_value, RATE_BINARY_SIZE);
        return snprintf(text_value, text_size, "%.3f",
                        UINT64_TO_DOUBLE(ntoh64(val_u64)));

      case PAYLOAD_BYTES:
        memcpy(&val_u64, bin_value, RATE_BINARY_SIZE);
        return snprintf(text_value, text_size, ("%" PRIu64),
                        ntoh64(val_u64));
    }

    return -1;
}


/*
 * int cut(unsigned int field, char *text_value, size_t text_size, rwRec *rwrec);
 *
 * If "field" is 0, then returns the number of supported fields.
 *
 * If "text_value" is NULL and "rwrec" is NULL, returns the size of
 * the buffer needed to store the title for this field.
 *
 * If "text_value" is NULL and "rwrec" is non-NULL, returns the size
 * of the buffer needed to store the value for this field.
 *
 * If "text_value" is non-NULL and "rwrec" is NULL, writes the title
 * into "text_value", and returns the number of bytes needed to store
 * that string.
 *
 * If "text_value" is non-NULL and "rwrec" is non-NULL, writes the
 * value for that record into "text_value", and returns the number of
 * bytes needed to store that string.
 */
int cut(unsigned int field, char *text_value, size_t text_size, rwRec *rwrec)
{
    if ((NULL != rwrec) && (NULL != text_value)) {
        return recToText(rwrec, text_value, text_size, &field);
    }
    if (NULL != text_value) {
        return getTitle(text_value, text_size, &field);
    }
    if (NULL != rwrec) {
        /* textual column width */
        return RATE_TEXT_WIDTH;
    }

    assert(field >= 1);
    assert(field <= 3);

    return 1 + strlen(plugin_fields[(field - 1)].name);
}


/*
 *  int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec);
 *
 *  Arguments:
 *      f:        field to get the binary or textual value of
 *      bin_val:  buffer to fill with a binary value that rwsort will
 *                use for sorting; rwsort will not ask the plug-in for
 *                more data; this value should be everything rwsort
 *                needs to know.  rwsort will use memcmp() to compare
 *                values; the values should be in big-endian byte
 *                order.  THIS BUFFER MAY NOT BE WORD ALIGNED.
 *      rec:      SiLK flow record
 *  Returns:
 *      when 'f' is zero
 *        => the number of supported fields
 *      when 'f' is non-zero
 *        when 'bin_val' or 'rec' is NULL
 *          => the number of bytes required to hold the binary value
 *             of field 'f'
 *        when 'bin_val' and 'rec' are non-NULL
 *          => length of the binary value; uses 'rec' to compute the
 *             value of field 'f' and writes that binary value into
 *             'bin_val'.  ASSUMES 'bin_val' IS BIG ENOUGH!
 */
int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec)
{
    if (bin_value && rwrec) {
        return recToBin(rwrec, bin_value, &field);
    }

    /* return width of binary value */
    return RATE_BINARY_SIZE;
}


/*
 *  int uniq(unsigned int f, uint8_t bin_val, char *text_value, size_t text_size, rwRec *rec)
 *
 *  Arguments:
 *      f:        field to get the binary or textual value of
 *      bin_val:  buffer to fill with binary value, or buffer to pass
 *                in the binary value when converting to text.
 *      text_value: buffer to fill with the textual value
 *      text_size: length of the buffer 'text_value'
 *      rec:      SiLK flow record
 *  Returns:
 *      when 'f' is zero
 *        => the number of supported fields
 *      when 'f' is non-zero
 *        when 'bin_val' is NULL
 *          => the result of running cut(f, text_value, text_size, rwrec)
 *        when 'bin_val' is non-NULL
 *          when 'text_value' is non-NULL and 'rec' has any value
 *            => length of buffer necessary to hold the complete
 *               textual value; fills 'text_value' with the first
 *               'text_size'-1 bytes of converting the binary value in
 *               'bin_val', corresponding to field 'f', to text.
 *          when 'text_value' is NULL and 'text_size' has any value
 *            when 'rec' is non-NULL
 *              => length of the binary value; uses 'rec' to compute
 *                 the value of field 'f' and writes that binary value
 *                 into 'bin_val'.  ASSUMES 'bin_val' IS BIG ENOUGH!
 *             when 'rec' is NULL
 *               => the number of bytes required to hold the binary
 *                  value of field 'f'
 */
int uniq(
    unsigned int    field,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_size,
    rwRec          *rwrec)
{
    if (bin_value) {
        if (text_value) {
            return binToText(bin_value, text_value, text_size, &field);
        }

        if (rwrec) {
            /* return the binary value for the given rwrec */
            return recToBin(rwrec, bin_value, &field);
        }

        /* return width of binary value */
        return RATE_BINARY_SIZE;
    }

    assert(field >= 1);
    assert(field <= 3);

    /* handle the title and width of textual field */
    if (NULL != text_value) {
        return getTitle(text_value, text_size, &field);
    }
    if (NULL != rwrec) {
        /* textual column width */
        return RATE_TEXT_WIDTH;
    }
    /* title width */
    return 1 + strlen(plugin_fields[(field - 1)].name);
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
